/*
 * spp.c
 *
 * Hardware level routines used for "quick-and-dirty" style in-circuit
 * programmers like Derren Crome's EPE programmer.  This driver includes 
 * support for voltage control and reading the PIC.  The PPSETUP value
 * can be used to choose either a parallel port or serial port version.
 * If the parallel port is selected it can be used as a replacement for
 * the EPE send program.
 *
 * Revision history:
 *
 * 03-Apr-1998: V-0.1; Created from pphw.c
 * 20-Apr-1998: V-0.2; Changed to use serial port (SERPIC)
 * 25-Apr-1998: V-0.3; Changed to use both parallel and serial ports
 *
 * Copyright (C) 1998 David Tait.  All rights reserved.
 * Permission is granted to use, modify, or redistribute this software
 * so long as it is not sold or exploited for profit.
 *
 * THIS SOFTWARE IS PROVIDED AS IS AND WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include "spp.h"
#include "hex.h"

#define PARALLEL 0
#define SERIAL   1

int base;
int port_type = PARALLEL;
int invert = 0;

int d_reg;              /* data reg */
int s_reg;              /* status reg */
int c_reg;              /* control reg */
int d_bits;

int mcr;                /* modem control reg */
int msr;                /* modem status reg */
int lcr;                /* line control reg */
int mcr_bits;
int lcr_bits;

int tdly = TDLY;
int dumpfmt = INHX8M;

void assert(int update)
{
    if ( update ) {
	if ( port_type == SERIAL ) {
	    outportb(mcr,mcr_bits);
	    outportb(lcr,lcr_bits);
	} else
	    outportb(d_reg,d_bits);
    }
}

unsigned inbit(void)
{
    if ( port_type == SERIAL )
	return ~inportb(msr)&SIN; 
    else
	return ~inportb(s_reg)&PIN;
}

void clkhi(int update)
{
    if ( port_type == SERIAL ) {
	if ( invert )
	    mcr_bits &= ~SCLK;
	else
	    mcr_bits |= SCLK;
    } else {
	if ( invert )
	    d_bits &= ~PCLK;
	else
	    d_bits |= PCLK;
    }
    assert(update);
}

void clklo(int update)
{
    if ( port_type == SERIAL ) {
	if ( invert )
	    mcr_bits |= SCLK;
	else
	    mcr_bits &= ~SCLK;
    } else {
	if ( invert )
	    d_bits |= PCLK;
	else
	    d_bits &= ~PCLK;
    }
    assert(update);
}

void outhi(int update)
{
    if ( port_type == SERIAL ) {
	if ( invert )
	    mcr_bits &= ~SOUT;
	else
	    mcr_bits |= SOUT;
    } else {
	if ( invert )
	    d_bits &= ~POUT;
	else
	    d_bits |= POUT;
    }
    assert(update);

}

void outlo(int update)
{
    if ( port_type == SERIAL ) {
	if ( invert )
	    mcr_bits |= SOUT;
	else
	    mcr_bits &= ~SOUT;
    } else {
	if ( invert )
	    d_bits |= POUT;
	else
	    d_bits &= ~POUT;
    }
    assert(update);
}

void vppon(int update)
{
    if ( port_type == SERIAL )
	lcr_bits &= ~SVPP;
    else
	d_bits &= ~PVPP;
    assert(update);
}

void vppoff(int update)
{
    if ( port_type == SERIAL )
	lcr_bits |= SVPP;
    else
	d_bits |= PVPP;
    assert(update);
}


void ppdelay(void)                 /* get delay from environment */
{
    char *s;

    if ( (s = getenv("PPDELAY")) != NULL )
	if ( (tdly = atoi(s)) < 0 || tdly > 127 )
	    tdly = TDLY;
}
 

void run_mode(int mode)    
{
    if ( mode == GO ) {
	vppoff(1);              
	printf("Remove programming voltage (if applied) and ");
	printf("hit any key to reset PIC ...\n");
	getch();
	clkhi(0), outhi(0), vppon(1);   /*  PIC should now run */ 
    } else
	clklo(0), outlo(0), vppoff(1);
}


void prog_mode(void)
{
    run_mode(STOP);
    ms_delay(TICKS(25000));
    vppon(1);
}


int interact(int wait)          
{
    run_mode(STOP);
    if ( wait ) { 
	printf("Apply programming voltage ... ");
	printf("hit any key to continue (^C to exit)\n\n");
	if ( getch() == 3 )
	    return 1;
    }
    return 0;
}


static void clock_out(int bit)
{
    bit? outhi(0): outlo(0); clkhi(1);
    us_delay(tdly);
    clklo(1);
    us_delay(tdly);
    outlo(1);
}


void out_word(int w)
{
    int b;

    clock_out(0);
    for ( b=0; b<14; ++b )
	clock_out(w&(1<<b));
    clock_out(0);
}


static int clock_in(void)
{
    clkhi(1);
    us_delay(tdly);
    clklo(1);
    us_delay(tdly);
    return inbit()? 1: 0;
}


int in_word(void)
{
    int b, w;

    (void) clock_in();
    for ( w=0, b=0; b<14; ++b )
	w += clock_in()<<b;
    (void) clock_in();

    return w;
}


void command(int cmd)
{
    int b;

    outlo(1);
    us_delay(tdly);
    for ( b=0; b<6; ++b )
	clock_out(cmd&(1<<b));
    outhi(1);
    us_delay(tdly);
}


void cleanup(void)
{
    reset_timer();
}

void debug(void)
{
    char *s;

    if ( atoi(getenv("PPDEBUG")) != 1 )
	return;
	
    printf("Hardware setup: %s port (%04X) with %s buffers (delay = %d)\n",
		       (port_type == SERIAL)? "Serial": "Parallel",
		       base,
		       (invert)? "inverting": "non-inverting",
		       tdly);
    printf("Debug mode entered ... (^C to exit) ");
    if ( getch() == 3 ) {
	printf("\n\n");
	exit(1);
    }
    printf("\nRemove PIC ... ");
    getch();
    for(;;) {
	vppon(0), clklo(0), outlo(1);
	us_delay(tdly);
	s = (inbit() == 0)? "OK" : "BAD";
	printf("\nRun, RB6 low, RB7 low (input %s) ... ",s);
	getch();
	clkhi(1);
	printf("\nRB6 high ... ");
	getch();
	clklo(0), outhi(1);
	us_delay(tdly);
	s = (inbit() != 0)? "OK" : "BAD";
	printf("\nRB7 high (input %s) ... ",s);
	getch();
	outlo(0), vppoff(1);
	printf("\nReset ... ");
	printf("\nStart over ... (^C to exit) ");
	if (getch() == 3) {
	    printf("\n\n");
	    exit(1);
	}
    }
}


    /* setup()
     *
     * Reads environment for port and tdly, initialises timer.
     * Returns -ve if port address is non-standard or 0 for all OK.
     *
     */

int setup(void)
{
    char *s;
    int port = 1;


    if ( (s = getenv("PPSETUP")) != NULL ) {
	port_type = (s[0] == 'P' || s[0] == 'p')? PARALLEL: SERIAL;
	invert = (s[1] == '0')? 0: 1;            
    }

    if ( port_type == SERIAL && (s = getenv("PPCOM")) != NULL )
	if ( (port = atoi(s)) < 1 || port > 4 )
	    port = 1;

    if ( port_type == PARALLEL ) {
	if ( (s = getenv("PPLPT")) != NULL )
	    if ( (port = atoi(s)) < 1 || port > 3 )
		port = 1;
	port += 4;
    }

    base = *(((int far *) 0x400) + port-1);  /* base address of port */

    switch ( base ) {           /* check port address is valid */
	case 0x2F8:
	case 0x3F8:
	case 0x2E8:
	case 0x3E8:
	case 0x3BC:
	case 0x278:
	case 0x378: break;
	   default: return -1;
    }
    
    ms_delay(0);                /* dummy call to setup CTC */

    if ( port_type == SERIAL ) {
	lcr = base + 3;
	mcr = base + 4;
	msr = base + 6;
	mcr_bits = inportb(mcr);
	lcr_bits = inportb(lcr);
    } else {
	d_reg = base;
	s_reg = base + 1;
	c_reg = base + 2;
	d_bits = 0xFF;
    }

    run_mode(STOP);
    ppdelay();
    debug();                    /* check if debugging requested */
    return 0;
}
